/*
	Copyright (C) 2006, 2007, 2011 Nicklas Nordborg

	This file is part of the Example Code Package for BASE.
	Available at http://baseplugins.thep.lu.se/
	BASE main site: http://base.thep.lu.se/
	
	This is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.
	
	The software is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.examples.plugins;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.DynamicSpotQuery;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.FloatParameterType;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.RequestInformation;
import net.sf.basedb.core.SpotBatcher;
import net.sf.basedb.core.Transformation;
import net.sf.basedb.core.VirtualColumn;

import net.sf.basedb.core.query.Dynamic;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Selects;
import net.sf.basedb.core.signal.SignalHandler;
import net.sf.basedb.core.signal.SignalTarget;
import net.sf.basedb.core.signal.ThreadSignalHandler;

import net.sf.basedb.core.plugin.AbstractAnalysisPlugin;
import net.sf.basedb.core.plugin.InteractivePlugin;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Set;


/**
	An example analysis plugin that creates a new bioassay set 
	by multiplying each intensitity value with the same factor, filtering
	out intensities below a cutoff value.
	
	@base.modified $Date: 2011-09-15 10:32:32 +0200 (Thu, 15 Sep 2011) $
	@author Nicklas
	@version 2.0
*/
public class ExampleAnalyzer 
	extends AbstractAnalysisPlugin
	implements InteractivePlugin, SignalTarget
{
	private static final Set<GuiContext> guiContexts = Collections.singleton(
		new GuiContext(Item.BIOASSAYSET, GuiContext.Type.ITEM)
	);
	
	// Job configuration is created by getConfigureJob()
	private RequestInformation configureJob;
	
	// Intensities below this value will be filtered out
	private PluginParameter<Float> cutoffParameter;
	
	// Multiply each intensity with this value
	private PluginParameter<Float> factorParameter;
	
	// So we can react to ABORT requests
	private ThreadSignalHandler signalHandler;
	
	/**
		Create a new plugin instance.
	*/
	public ExampleAnalyzer()
	{}
	
	/*
		From the Plugin interface
		-------------------------------------------
	*/
	@Override
	public boolean supportsConfigurations()
	{
		return false;
	}
	@Override
	public boolean requiresConfiguration()
	{
		return false;
	}
	@Override
	public void run(Request request, Response response, ProgressReporter progress)
	{
		// Ensure that the ABORT signal gets sent to this thread
		if (signalHandler != null) signalHandler.setWorkerThread(null);

		// Open a connection to the database
		// sc is set by init() method
		DbControl dc = sc.newDbControl();
		try
		{
			// Get job parameters
			float cutoff = (Float)job.getValue("cutoff");
			float factor = (Float)job.getValue("factor");
			BioAssaySet source = getSourceBioAssaySet(dc);
			int channels = source.getRawDataType().getChannels();
	      
			// Create transformation and child bioassay set
			Job j = Job.getById(dc, job.getId());
			Transformation t = source.newTransformation(j);
			t.setName("Multiply " + factor + "; filter < " + cutoff);
			BioAssaySet result = t.newProduct(null, "new", true);
			result.setName("After: multiply and filter");
			dc.saveItem(t);
			dc.saveItem(result);
	      
			// Get query for source data
			DynamicSpotQuery query = source.getSpotData();
	      
			// Query expressions for multiplication factor and cutoff
			Expression factorExpression = Expressions.aFloat(factor);
			Expression cutoffExpression = Expressions.aFloat(cutoff);
				      
			// Specify select columns
			query.select(Dynamic.select(VirtualColumn.COLUMN));
			query.select(Dynamic.select(VirtualColumn.POSITION));
			for (int ch = 1; ch <= channels; ch++)
			{
				Expression chExpression = Dynamic.column(VirtualColumn.channelRaw(ch));
				// Selects: ch * factor for each channel
				query.select(
					Selects.expression(Expressions.multiply(chExpression, factorExpression), "ch" + ch)
				);
				// Restricts: ch >= cutoff for each channel
				query.restrict(Restrictions.gteq(chExpression, cutoffExpression));
			}
	      
			// Create batcher and copy data
			SpotBatcher batcher = result.getSpotBatcher();
			// Check for the ABORT signal before a lengthy operation
			ThreadSignalHandler.checkInterrupted(); 
			int spotsCopied = batcher.insert(query);
			batcher.close();
			int spotsRemoved = source.getNumSpots() - spotsCopied;
	      
			// Commit the work
			dc.commit();
			response.setDone("Multiplied " + spotsCopied + " spots with factor " + factor + 
				"; removed " + spotsRemoved + " spots with intensity < " + cutoff);
		}
		catch (Throwable t)
		{
			// All exceptions must be catched and sent back 
			// using the response object
			response.setError(t.getMessage(), Arrays.asList(t));
		}
		finally
		{
			// IMPORTANT!!! Make sure opened connections are closed
			if (dc != null) dc.close();
		}
	}
	// -------------------------------------------
	
	/*
		From the InteractivePlugin interface
		-------------------------------------------
	*/
	/**
		Return a set containing the context [BIOASSAYSET, ITEM].
	*/
	@Override
	public Set<GuiContext> getGuiContexts()
	{
		return guiContexts;
	}
	/**
		Check if the item is a {@link BioAssaySet} and the logged in
		user has USE permission to the experiment. Also verify that
		the bioassay set has data stored in the database.
	*/
	@Override
	public String isInContext(GuiContext context, Object item)
	{
		String message = super.isInContext(context, item);
		if (message == null)
		{
			BioAssaySet bas = (BioAssaySet)item;
			if (bas.getNumSpots() <= 0)
			{
				message = "This plug-in requires spot data in the database";
			}
		}
		return message;
	}
	/**
		The {@link Request#COMMAND_CONFIGURE_PLUGIN} command will not ask
		for any parameters.
		<p>
		The {@link Request#COMMAND_CONFIGURE_JOB} command will ask for
		a bioassay set, a multiplcation factor and a cutoff value.
	*/
	@Override
	public RequestInformation getRequestInformation(GuiContext context, String command)
		throws BaseException
	{
		RequestInformation requestInformation = null;
		if (command.equals(Request.COMMAND_CONFIGURE_JOB))
		{
			requestInformation = getConfigureJob(context);
		}
		return requestInformation;
	}
	/**
		Store configuration settings for {@link Request#COMMAND_CONFIGURE_PLUGIN} and
		{@link Request#COMMAND_CONFIGURE_JOB}.
	*/
	@Override
	public void configure(GuiContext context, Request request, Response response)
	{
		String command = request.getCommand();
		try
		{
			if (command.equals(Request.COMMAND_CONFIGURE_JOB))
			{
				RequestInformation ri = getConfigureJob(context);
				List<Throwable> errors = 
					validateRequestParameters(ri.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size()+" invalid parameter(s) were found in the request", errors);
					return;
				}

				// Source bioassay set
				storeValue(job, request, ri.getParameter(SOURCE_BIOASSAYSET));
								
				// Other options
				storeValue(job, request, factorParameter);
				storeValue(job, request, cutoffParameter);
				BioAssaySet bioAssaySet = (BioAssaySet)job.getValue(SOURCE_BIOASSAYSET);
				response.setSuggestedJobName("Example analysis on bioassay set '" + 
						bioAssaySet.getName() + "'");
				response.setDone("Job configuration complete", Job.ExecutionTime.SHORT);
				// TODO - maybe check file size to make a better estimate
			}
		}
		catch (Throwable ex)
		{
			response.setError(ex.getMessage(), Arrays.asList(ex));
		}
	}
	// -------------------------------------------
	/*
		From the SignalTarget interface
		-------------------------------------------
	*/
	@Override
	public SignalHandler getSignalHandler()
	{
		signalHandler = new ThreadSignalHandler();
		return signalHandler;
	}
	// -------------------------------------------


	private RequestInformation getConfigureJob(GuiContext context)
	{
		DbControl dc = null;
		try
		{
			if (configureJob == null)
			{
				dc = sc.newDbControl();
				BioAssaySet bas = getCurrentBioAssaySet(dc);
				// Create parameter list and request informaion
				List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();
				
				// Source bioassay set parameter
				parameters.add(getSourceBioAssaySetParameter(null, null));		
	
				// Other parameters
				// Create multiplication factor parameter
				factorParameter = new PluginParameter<Float>(
					"factor",
					"Multiplication factor",
					"The factor to multiply each spot intensity with.",
					new FloatParameterType(0.0f, 100.0f, 1.0f, true)
				);
	
				// Create cutoff parameter
				cutoffParameter = new PluginParameter<Float>(
					"cutoff",
					"Cut-off value",
					"Original intensities falling below this value are filtered out.",
					new FloatParameterType(null, null, 0.0f, true)
				);
				parameters.add(factorParameter);
				parameters.add(cutoffParameter);
				
				configureJob = new RequestInformation
				(
					Request.COMMAND_CONFIGURE_JOB,
					"Specify multiplication factor and cutoff",
					"An example analysis plugin that creates a new bioassay set " + 
					"by multiplying each intensitity value with the same factor, " +
					"filtering out intensities below a cutoff value.",
					parameters
				);
			}
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return configureJob;
	}
}
